/* -*- C++ -*- Time-stamp: <10 Mar 01 18:06:49 Michael Bischoff> */

/*****************************************************************************/
/*									     */
/*									     */
/*	QMatchIt version 1.0 -- file game.cpp				     */
/*									     */
/*	Include file for the field base class of QMatchIt		     */
/*	written by Michael Bischoff (Michael.Bischoff@gmx.net)		     */
/*	November-2000							     */
/*	see COPYRIGHT.qmatchit for Copyright details			     */
/*									     */
/*									     */
/*****************************************************************************/

#include "qmatchit.h"		// global constants
#include "fields.h"		// field type classes
#include "game.h"

#include <iostream.h>
#define UNSAFE_STOPS_IF_BLOCK_AWAY

game::game() {
    rows = new field **[MAXX];
    for (int x = 0; x < MAXX; ++x) {
	rows[x] = new field *[MAXY];
	for (int y = 0; y < MAXY; ++y)
	    rows[x][y] = 0;
    }

    engaged_with = 0;	

    cursor = field::newField(RTTI_CURSOR, 0, 0);
    cursor_x = cursor_y = 0;
    nextpoints = 1;
    score = 0;
    lv_title = 0;
    lv_desc = 0;
}

void game::disengageCursor() {
    if (engaged_with) {
	engaged_with = 0;
	cursor->setState(field::state_normal);
    }
}

void game::engageCursor(field *fp) {
    engaged_with = fp;
    cursor->setState(field::state_anim);
}

void game::clear() {
    disengageCursor();
    cursor_x = cursor_y = 0;
    cursor->setXY(0, 0);
    for (int x = 0; x < MAXX; ++x) {
	for (int y = 0; y < MAXY; ++y) {
	    if (rows[x][y]) {
		delete rows[x][y];
		rows[x][y] = 0;
	    }
	}
    }
    score = 0;
    nextpoints = 1;
}

void game::deleteTexts() {
    if (lv_title) {
	delete lv_title;
	lv_title = 0;
    }
    if (lv_desc) {
	delete lv_desc;
	lv_desc = 0;
    }
}

game::~game() {
    clear();	// clear all field objects
    deleteTexts();
    delete cursor;
    for (int x = 0; x < MAXX; ++x)
	delete rows[x];
    delete rows;
}


// create a game from level data
// must have called clear() before
void game::setRows(const char **dat) {
    const char *p;

    for (int y = 0; y < MAXY && *dat; ++y, ++dat) {
	int x;
	for (x = 0, p = *dat; *p && x < MAXX; ++x, ++p) {
	    if (*p != '.' && *p != ' ') {
		rows[x][y] = field::newField(*p, x, y);
		// consistency checks
		if (rows[x][y]) {
		    if (rows[x][y]->ascii() != *p)
			cout << "Warning! internal error: ascii " << (int)(*p) << " gives " << rows[x][y]->ascii() << endl;
		} else {
		    cout << "Warning! Field of type " << *p << " ignored!" << endl;
		}
	    }
	}
    }
}

// check for actions if cell is cleared
void game::transitionToEmpty(int x, int y) {
    rows[x][y] = 0;
    field *fp = rows[x][y+1];
    if (fp && fp->rtti() == RTTI_VANISHER && fp->stateIs(field::state_normal)) {
	// kill block now
	fp->startDestruction();
    }
}

void game::changeAllMagnets() {
    for (int y = 0; y < MAXY; ++y) {
	for (int x = 0; x < MAXX; ++x) {
	    field *fp = rows[x][y];
	    if (fp && fp->rtti() == RTTI_MAGNET) {
		static_cast<f_magnet *>(fp)->invert();
	    }
	}
    }
}

// rows[x][y] is guaranteed to hold a block
void game::moveBlock(int x, int y, int x2, int y2) {
    field *fp = rows[x][y];
    fp->idle_timer = 0;
    rows[x2][y2] = fp;
    rows[x][y] = 0;
    if (fp == engaged_with) {
	/* move down cursor as well */
	cursor_x = x2;
	cursor_y = y2;
	cursor->setXY(cursor_x, cursor_y);
    }
    fp->setXY(x2, y2);
}

// wrapper to avoid using class field from mainwindow directly
void game::advance_all() {
    field::advance_all();
}

// animate() performs most of the automatic actions on the playground:
// 1. remove any dead objects (vanished boxes or bombs, unsolid ground)
// 1a. this may trigger the destruction of any blocks they're on.
// 2. animation of the QCanvasItem sprites
// 3. trigger (falling) movement of boxes
// 4. trigger the destruction of matching blocks which have not been
//    moved in step 3. (because this would explode boxes when other
//    are falling down beside them)


void game::animate() {
    if (!rows)
	return;	// savety check

    int y;
    for (y = MAXY-1; y >= 0; --y) {
	for (int x = 0; x < MAXX; ++x) {
	    field *fp = rows[x][y];
	    if (fp) {
		fp->idle_timer++;
#ifdef UNSAFE_STOPS_IF_BLOCK_AWAY
		if (y && fp->rtti() == RTTI_UNSAFE && !rows[x][y-1]) {
		    fp->stopDestruction();
		}
#endif
		// remove dead blocks
		if (fp->stateIs(field::state_dead)) {
		    delete fp;
		    rows[x][y] = 0;
		    transitionToEmpty(x, y);
		}
	    }
	}
    }
    field::advance_all();

    // check for items over empty spaces or unsafe ground
    for (y = 1; y < MAXY-1; ++y) {
	for (int x = 1; x < MAXX-1; ++x) {
	    //cout << "check " << x << y << endl;
	    field *fp = rows[x][y];
	    if (fp && fp->idle_timer) {
		if (fp->rtti() == RTTI_BALLOON && !rows[x][y-1]) {
		    /* move up balloon */
		    moveBlock(x, y, x, y-1);
		    continue;
		}
		if (fp->rtti() == RTTI_LIFT) {
		    field *box;
		    if ((box = rows[x][y-1])) {
			if (y > 2 && !rows[x][y-2] &&
			    box->rtti() == RTTI_BOX &&
			    box->idle_timer && !heldByMagnet(x, y-1)) {
			    /* move up lift and box on it */
			    moveBlock(x, y-1, x, y-2);
			    moveBlock(x, y, x, y-1);
			    continue;
			}
		    } else {
			// field above is empty: only type 1 lift goes up
			if (fp->getSubtype() == 1) {
			    moveBlock(x, y, x, y-1);
			    continue;
			}
			if (fp->getSubtype() == 2 && !rows[x][y+1]) {
			    // lift type 2 falls down again
			    moveBlock(x, y, x, y+1);
			    continue;
			}
		    }
		}
		if (fp->feelsGravity() && fp->stateIs(field::state_normal)) {
		    if (!rows[x][y+1]) {
			// box over empty space
			// check for magnets! (not for gravity blocks)
			if (fp->feelsMagnets() && heldByMagnet(x, y))
			    continue;
			// yupp, definitely falling down
			moveBlock(x, y, x, y+1);
			continue;
		    } else if (rows[x][y+1]->rtti() == RTTI_UNSAFE) {
			/* destruct item */
			rows[x][y+1]->startDestruction();
		    }
		}
	    }
	}
    }

    for (y = 1; y < MAXY-1; ++y) {
	for (int x = 1; x < MAXX-1; ++x) {
	    //cout << "check " << x << y << endl;
	    field *fp = rows[x][y];
	    if (fp && fp->idle_timer && fp->rtti() == RTTI_BOX) {
		int this_rtti = rows[x][y+1] ? rows[x][y+1]->rtti() : 0;
		if (this_rtti == RTTI_SAVESPOT) {
		    ;	// save in any case
		} else if (this_rtti == RTTI_BOMB) {
		    // definitely killed
		    destruct(rows[x][y+1]);
		    destruct(fp);
		} else if (fp->getSubtype()) {
		    /* maybe is an item */
		    // cout << "check " << x << ", " << y << endl;
		    checkForDestruction(x, y,  0,  1);
		    checkForDestruction(x, y,  1,  0);
		    checkForDestruction(x, y,  0, -1);
		    checkForDestruction(x, y, -1,  0);
		}
	    }
	}
    }

    // if (!advanceLevel)
	// advanceLevel = levelIsSolved();
}

bool game::levelIsSolved() {
    //cout << "Test starts" << endl;
    // maybe the level is solved now: initiate message then
    for (int y = 0; y < MAXY; ++y) {
	for (int x = 0; x < MAXX; ++x) {
	    field *fp = rows[x][y];
	    if (fp) {
		if (!fp->idle_timer)
		    return false;	// recently changed
		if (fp->rtti() == RTTI_BOX && fp->getSubtype())
		    return false;	// boxes are remaining
	    }
	}
    }
    //cout << "Test ends with level solved!" << endl;
    return true;
}			

// rows[x][y] is a box. check for neighbors of the same type

void game::checkForDestruction(int x, int y, int dx, int dy) {
    field *fp2 = rows[x+dx][y+dy];
    if (fp2 && fp2->rtti() == RTTI_BOX && fp2->idle_timer) {
	field *fp = rows[x][y];
	if (fp && fp2->getSubtype() == fp->getSubtype())
	    destruct(rows[x][y]);
    }
}

// central point to start removing a tile
void game::destruct(field *p) {
    if (p->stateIs(field::state_normal)) {
	// kill this tile!
	if (p->rtti() == RTTI_BOX && p->getSubtype()) {
	    // this is a normal box: score for it!
	    // implement quadratic scoring: 4 pts for 2 boxes, 9 for 3 etc...
	    //cout << "adding " << nextpoints << " to score" << endl;
	    score += nextpoints;
	    nextpoints += 2;
	    //cout << "score=" << score << ", nextpoints=" << nextpoints << endl;
	}
	p->startDestruction();

	if (engaged_with == p) {
	    disengageCursor();
	}
    }
}


void game::gameKey(enum key_dir key) {
    nextpoints = 1;	// reset score bonus

    if (!engaged_with) {
	/* free movement when no block connected */
	switch (key) {
	case Key_Up:
	    if (cursor_y) {
		--cursor_y;
		cursor->setXY(cursor_x, cursor_y);
	    }
	    break;
	case Key_Left:
	    if (cursor_x) {
		--cursor_x;
		cursor->setXY(cursor_x, cursor_y);
	    }
	    break;
	case Key_Down:
	    if (cursor_y < MAXY-1) {
		++cursor_y;
		cursor->setXY(cursor_x, cursor_y);
	    }
	    break;
	case Key_Right:
	    if (cursor_x < MAXX-1) {
		++cursor_x;
		cursor->setXY(cursor_x, cursor_y);
	    }
	    break;
	case Key_Select:
	    /* grab block if available */
	    field *fp = rows[cursor_x][cursor_y];
	    if (fp && fp->idle_timer
		&& fp->isSelectable()
		&& fp->stateIs(field::state_normal)) {
		switch (fp->rtti()) {
		case RTTI_BOX:
		    engageCursor(fp);
		    break;
		case RTTI_INVERTER:
		    // change all magnets, use up this object
		    fp->startDestruction();
		    changeAllMagnets();
		    break;
		case RTTI_REMOVABLE:
		    // just remove it, no further action
		    fp->startDestruction();
		    break;
		default:
		    break;
		}
	    }
	    break;
	}
    } else {
	/* active block */
	int this_rtti = rows[cursor_x][cursor_y+1] ?
	    rows[cursor_x][cursor_y+1]->rtti() : 0;
	field *fp = rows[cursor_x][cursor_y];
	switch (key) {
	case Key_Select:
	    /* drop block */
	    disengageCursor();
	    break;
	case Key_Down:
	    if (fp->idle_timer < 2)
		return;		// just arrived here
	    
	    if (fp->feelsMagnets()
		&& heldByMagnet(cursor_x, cursor_y))
		return;

	    // down direction only possible if on teleporter
	    // find matching teleporter
	    if (this_rtti == RTTI_TELEPORT) {
		int this_subtype = rows[cursor_x][cursor_y+1]->getSubtype();
		for (int y = 0; y < MAXY-1; ++y) {
		    for (int x = 0; x < MAXX; ++x) {
			if (rows[x][y+1] &&
			    rows[x][y+1]->rtti() == RTTI_TELEPORT &&
			    rows[x][y+1]->getSubtype() == this_subtype &&
			    !rows[x][y] &&	
			    (x != cursor_x || y != cursor_y)) {
			    // yeah, it's the other teleporter
			    moveBlock(cursor_x, cursor_y, x, y);
			    return;
			}
		    }
		}
	    }
	    break;
	case Key_Right:
	    if (fp->idle_timer < 2)
		return;		// just arrived here
	    
	    if (fp->feelsMagnets()
		&& heldByMagnet(cursor_x, cursor_y))
		return;

	    // may not move if falling (free space below it)
	    if (this_rtti == 0)
		return;		// no way ;-)

	    // check for wrong way
	    if (this_rtti == RTTI_ONEWAY
		&& rows[cursor_x][cursor_y+1]->getSubtype() == f_oneway::oneway_left)
		return;		// no way ;-)

	    if (cursor_x < MAXX-1) {
		if (!rows[cursor_x+1][cursor_y]) {	
		    /* yeah, move it! */
		    moveBlock(cursor_x, cursor_y, cursor_x+1, cursor_y);
		    transitionToEmpty(cursor_x-1, cursor_y);
		}
	    }
	    break;
	case Key_Left:
	    if (fp->idle_timer < 2)
		return;		// just arrived here
	    
	    if (fp->feelsMagnets()
		&& heldByMagnet(cursor_x, cursor_y))
		return;

	    // may not move if falling (free space below it)
	    if (this_rtti == 0)
		return;		// no way ;-)

	    // check for wrong way
	    if (this_rtti == RTTI_ONEWAY
		&& rows[cursor_x][cursor_y+1]->getSubtype() == f_oneway::oneway_right)
		return;		// no way ;-)

	    if (cursor_x > 0) {
		if (!rows[cursor_x-1][cursor_y]) {
		    /* yeah, move it! */
		    moveBlock(cursor_x, cursor_y, cursor_x-1, cursor_y);
		    transitionToEmpty(cursor_x+1, cursor_y);
		}
	    }
	    break;
	case Key_Up:
	    break;
	}
    }
}

#define TEST(dx, dy, dir)	if ((fp = rows[x+(dx)][y+(dy)]) && fp->rtti() == RTTI_MAGNET && fp->getSubtype() == f_magnet::dir) return true


bool game::heldByMagnet(int x, int y) {
    field *fp;
    TEST( 0,  1, magnet_up);
    TEST( 0, -1, magnet_down);
    TEST( 1,  0, magnet_left);
    TEST(-1,  0, magnet_right);
    return false;
}

void game::editKey(int ascii) {            // keyPress in edit mode
    if (engaged_with)
	disengageCursor();
    if (ascii < 0x20) {
	gameKey(static_cast<enum key_dir>(ascii));	// same to do
	return;
    }
    // printable character: delete field and perhaps create new one
    if (rows[cursor_x][cursor_y]) {
	delete rows[cursor_x][cursor_y];
	rows[cursor_x][cursor_y] = 0;
    }
    rows[cursor_x][cursor_y] = field::newField(ascii, cursor_x, cursor_y);
}

// level I/O is mainly C code
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

bool game::loadFromFile(const char *lv_filename, int level) {
    char *p, match[30];
    char line[2000];
    bool found = FALSE;
    FILE *fp;
    static char level_x[MAXY+1][MAXX+10];
    static char *level_t[MAXY+1];
    sprintf(match, "Level %d:\n", level);

    deleteTexts();

    if (!(fp = fopen(lv_filename, "r"))) {
	fprintf(stderr, "Cannot read from %s\n", lv_filename);
	return false;
    }
    while (fgets(line, sizeof(line), fp)) {
	if (!strcmp(line, match)) {
	    found = TRUE;
	    break;
	}
    }
    if (!found) {
	fprintf(stderr, "Cannot find level %d in %s\n", level, lv_filename);
	return false;
    }
    fgets(line, sizeof(line), fp);
    if ((p = strchr(line, '\n')))
	*p = '\0';
    lv_title = new char [strlen(line) + 1];
    strcpy(lv_title, line);
    fgets(line, sizeof(line), fp);
    lv_desc = new char [strlen(line) + 1];
    strcpy(lv_desc, line);
    
    memset(level_x, 0, sizeof(level_x));
    int y;
    for (y = 0; y < MAXY; ++y) {
	level_t[y] = level_x[y];
	if (!(fgets(level_x[y], MAXX+2, fp)))
	    break;
	if (!strncmp(level_x[y], "Level ", 6))
	    break;
	char *s = strchr(level_x[y], '\n');
	if (s)
	    *s = '\0';
    }
    level_t[y] = 0;
    fclose(fp);
    clear();
    setRows((const char **)level_t);
    return true;
}

// save the current game
void game::save(const char *filename) {
    FILE *fp;
    if (!(fp = fopen(filename, "w"))) {
	fprintf(stderr, "Cannot save to %s\n", filename);
	return;
    }
    for (int y = 0; y < MAXY; ++y) {
	char line[MAXX+1];
	for (int x = 0; x < MAXX; ++x) {
	    line[x] = rows[x][y] ? rows[x][y]->ascii() : '.';
	}
	line[MAXX] = '\0';
	fprintf(fp, "%s\n", line);
    }
    fclose(fp);
}

// load a single level
void game::load(const char *filename) {
    FILE *fp;
    static char level_x[MAXY+1][MAXX+10];
    static char *level_t[MAXY+1];
    if (!(fp = fopen(filename, "r"))) {
	fprintf(stderr, "Cannot read from %s\n", filename);
	return;
    }
    memset(level_x, 0, sizeof(level_x));
    int y;
    for (y = 0; y < MAXY; ++y) {
	level_t[y] = level_x[y];
	if (!(fgets(level_x[y], MAXX+2, fp)))
	    break;
	char *s = strchr(level_x[y], '\n');
	if (s)
	    *s = '\0';
    }
    level_t[y] = 0;
    fclose(fp);
    clear();
    setRows((const char **)level_t);
}
